{   This file is a part of Witch7DRL, the roguelike game.
    Copyright (C) 2011 Konovod Andrey (kipar@front.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/
}
unit uMain;

{$mode objfpc}{$H+}

interface

uses
  Windows, Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  StdCtrls, uClass, uOpenGLGraph;

type

  { TForm1 }

  TForm1 = class(TForm)
    Memo1: TMemo;
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormResize(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure Image1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Timer1Timer(Sender: TObject);
  private
    { private declarations }
  public
    { public declarations }
  end; 

var
  Form1: TForm1;
  Cache: TBitmap;
  DoProcess: boolean;
  ScrollX, ScrollY: Integer;

procedure Log(S: string);

implementation

uses uManual;


procedure Log(S: string);
begin
 if USELOG then Form1.Memo1.Lines.Add(S);
end;

{$R *.lfm}

{ TForm1 }

procedure TForm1.FormShow(Sender: TObject);
begin
  fmManual.Show;
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  case Key of
    vk_space: DoProcess := not DoProcess;
    Ord('G'): THEGAME.GenLevel;
    Ord('D'): THEGAME.Player.DrawMode := not THEGAME.Player.DrawMode;
    Ord('F'): THEGAME.Player.InsertSource(Fire);
    Ord('W'): THEGAME.Player.InsertSource(Water);
    Ord('A'): THEGAME.Player.InsertSource(Air);
    Ord('E'): THEGAME.Player.InsertSource(Earth);
  end;
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  ResizeGL(Self);
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  randomize;
  THEGAME := TGame.Create;
  Cache := TBitmap.Create;
  Cache.Width := MAXX;
  Cache.Height := MAXY;
  DoProcess := false;
  if not USELOG then Memo1.Visible := false;
  InitGL(Self);
  //THEGAME.GenLevel;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FreeGL(Self);
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState
  );
begin
  case Key of
    VK_SUBTRACT:
      begin
        THEGAME.VIEWSCALE := THEGAME.VIEWSCALE*1.1;
        ResizeGL(Self);
      end;
    VK_ADD:
    begin
      THEGAME.VIEWSCALE := THEGAME.VIEWSCALE*0.9;
      ResizeGL(Self);
    end;
    Ord('S'): THEGAME.ProcessTime;
  end;
end;

procedure TForm1.Image1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  RX, RY: Extended;
begin
  RX := X / ClientWidth * MAXX*THEGAME.VIEWSCALE + THEGAME.Player.X - MAXX*THEGAME.VIEWSCALE/2;
  RY := Y / ClientHeight * MAXY*THEGAME.VIEWSCALE + THEGAME.Player.Y - MAXY*THEGAME.VIEWSCALE/2;
  if Button = mbLeft then
  begin
    THEGAME.Player.SetDirTo(RX, RY);
    DoProcess := False;
  end;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  Timer1.Enabled := false;
  if DoProcess then
    THEGAME.ProcessTime;
  with THEGAME.Player do
    if abs(TargetX - X)+abs(TargetY - Y)<1e-3 then
      DoProcess := false;
  UpdateFrame;
  Timer1.Enabled := True;
end;

end.

