{   This file is a part of Witch7DRL, the roguelike game.
    Copyright (C) 2011 Konovod Andrey (kipar@front.ru).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see http://www.gnu.org/licenses/
}
unit uOpenGLGraph;

{$mode objfpc}{$H+}

interface

uses
  Windows, Classes, SysUtils, GL, Forms;

type
  TLineEffect = (Scratch, Current, Target, Ready, Active, WideActive);

  TOpenGLColor = record
    R, G, B, A: GLfloat;
  end;

procedure InitGL(Form: TForm);
procedure ResizeGL(Form: TForm);
procedure FreeGL(Form: TForm);
procedure UpdateFrame;
procedure DrawObject(L: integer; X, Y: GLfloat);
procedure DrawLine(X1, Y1, X2, Y2: GLfloat; Color1, Color2: TOpenGLColor;
  Effect: TLineEffect);
procedure DrawRect(X1, Y1, X2, Y2: GLfloat; const Color: TOpenGLColor);
procedure TexturedRect(X1, Y1, X2, Y2: GLfloat; Tex: integer; TexScale: extended);
procedure SetColor(const C: TOpenGLColor);
procedure SetRandomColor(const C, Delta: TOpenGLColor);
function MixColor(C1, C2: TOpenGLColor; Rate: GLfloat): TOpenGLColor;

procedure Circle(R: GLfloat; CenterColor, EdgeColor, VarColor: TOpenGLColor);

function CLR(aR, aG, aB, aA: GLfloat): TOpenGLColor;
function InvCLR(C: TOpenGLColor): TOpenGLColor;

operator = (C1, C2: TOpenGLColor): boolean;

procedure SetScreenGamma(C: TOpenGLColor);


const
  BLACK: TOpenGLColor = (R: 0; G: 0; B: 0; A: 1);
  GRAY: TOpenGLColor = (R: 0.5; G: 0.5; B: 0.5; A: 1);
  WHITE: TOpenGLColor = (R: 1; G: 1; B: 1; A: 1);
  RED: TOpenGLColor = (R: 1; G: 0; B: 0; A: 1);
  GREEN: TOpenGLColor = (R: 0; G: 1; B: 0; A: 1);
  BLUE: TOpenGLColor = (R: 0; G: 0; B: 1; A: 1);
  SKYBLUE: TOpenGLColor = (R: 166 / 256; G: 202 / 256; B: 240 / 256; A: 1);
  TRANSPARENT: TOpenGLColor = (R: 0; G: 0; B: 0; A: 0);

  PURPLE: TOpenGLColor = (R: 1; G: 0; B: 1; A: 1);
  YELLOW: TOpenGLColor = (R: 1; G: 1; B: 0; A: 1);

  WARRIORSIGHT: TOpenGLColor = (R: 1; G: 0; B: 0; A: 0.1);
  DEMONSIGHT: TOpenGLColor = (R: 0.5; G: 0; B: 0; A: 0.3);
  ZOMBIESIGHT: TOpenGLColor = (R: 0; G: 0; B: 0; A: 0.8);
  HASTESIGHT: TOpenGLColor = (R: 0; G: 1; B: 0.5; A: 0.1);

  TEX_GROUND=1;
  TEX_WALL=2;
  TEX_WITCH=3;
  TEX_WARRIOR=4;


  FLOOR_COLOR=0.4;

  GAMMA_DECR = 0.1;

var
  CurGamma: TOpenGLColor;



implementation

uses uClass;

var
  DC:  HDC;
  hrc: HGLRC;

{ЭТУ ФУНКЦИЮ Я ВЗЯЛ ИЗ КНИГИ М.КРАСНОВА
 "OPENGL-ГРАФИКА В ПРОЕКТАХ DELPHI"}
procedure SetDCPixelFormat(hdc: HDC);
var
  pfd: TPixelFormatDescriptor;
  nPixelFormat: integer;
begin
  FillChar(pfd, SizeOf(pfd), 0);
  pfd.dwFlags  := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or PFD_DOUBLEBUFFER;
  nPixelFormat := ChoosePixelFormat(hdc, @pfd);
  SetPixelFormat(hdc, nPixelFormat, @pfd);
end;

{ЭТУ ФУНКЦИЮ Я ВЗЯЛ ИЗ КНИГИ М.КРАСНОВА
 "OPENGL-ГРАФИКА В ПРОЕКТАХ DELPHI"}
procedure PrepareImage(filename: string; Alpha: boolean);
const
  szh = SizeOf(TBitmapFileHeader);
  szi = SizeOf(TBitmapInfoHeader);
type
  TRGB = record
    r, g, b: GLUbyte;
  end;

  TRGBA = record
    r1, g1, b1, a1: GLUbyte;
  end;
  TWrap  = array [0..100000000] of TRGB;
  PWrap  = ^TWrap;
  TWrapA = array [0..100000000] of TRGBA;
  PWrapA = ^TWrapA;
var
  BmpFile: file;
  bfh:     TBitmapFileHeader;
  bmi:     TBitmapInfoHeader;
  x, size: GLint;
  sWidth, tHeight: GLSizei;
  Data:    PWrap;
  dataA:   PWrapA;
  ar, ag, ab: integer;
begin
  AssignFile(BmpFile, FileName);
  Reset(BmpFile, 1);
  Size := FileSize(BmpFile) - szh - szi;
  Blockread(BmpFile, bfh, szh);
  BlockRead(BmpFile, bmi, szi);
  if Bfh.bfType <> $4D42 then
    Exit;
  sWidth  := bmi.biWidth;
  tHeight := bmi.biHeight;
  GetMem(Data, Size);
  GetMem(DataA, size * 4 div 3);
  BlockRead(BmpFile, Data^, Size);
  ar := Data^[0].r;
  ag := Data^[0].g;
  ab := Data^[0].b;
  for x := 0 to sWidth * tHeight - 1 do
    with Data^[x], dataA^[x] do
    begin
      b1 := r;
      r1 := b;
      g1 := g;
      //      a1 := 255;
      if (r = ar) and (g = ag) and (b = ab) then
        a1 := 1
      else
      if abs(r - ar) + abs(g - ag) + abs(b - ab) < 100 then
        a1 := 128
      else
        a1 := 255;
    end;
  glEnable(GL_BLEND);
  if Alpha then
    glTexImage2D(GL_TEXTURE_2D, 0, 4, sWidth, tHeight, 0,
      GL_RGBA, GL_UNSIGNED_BYTE, dataA)
  else
    glTexImage2D(GL_TEXTURE_2D, 0, 3, sWidth, tHeight, 0,
      GL_RGBA, GL_UNSIGNED_BYTE, dataA);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  Freemem(Data);

end;


procedure InitTextures;
var
  I: integer;
  S: string;
begin
  i := 1;
  repeat
    s := '.\tex\' + IntToStr(i) + '.bmp';
    if FileExists(s) then
    begin
      glBindTexture(GL_TEXTURE_2D, i);
      PrepareImage(s, (i > 2));
    end;
    Inc(i)
  until i = 100;
end;

procedure InitLists;
var
  phi: GLfloat;
begin
  glNewList(MONSTERLIST, GL_COMPILE);
  Circle(MINDIST / 2, BLACK, GREEN, BLACK);
  glEndList;

  glNewList(DEMONLIST, GL_COMPILE);
  Circle(MINDIST, RED, BLACK, BLACK);
  SetColor(RED);
  glLineWidth(2);
  glBegin(GL_LINE_STRIP);
  phi := 0;
  repeat
    glVertex2f(cos(phi) * MINDIST, sin(phi) * MINDIST);
    phi := phi + 4*pi / 5
  until phi > 4 * pi + 0.1;
  glEnd;
  glEndList;

  glNewList(GOLEMLIST, GL_COMPILE);
  Circle(MINDIST, YELLOW, BLACK, BLACK);
  SetColor(BLACK);
  glLineWidth(2);
  glBegin(GL_LINE_STRIP);
  phi := 0;
  repeat
    glVertex2f(cos(phi) * MINDIST, sin(phi) * MINDIST);
    phi := phi + 4*pi / 6
  until phi > 4 * pi + 0.1;
  phi := pi/3;
  repeat
    glVertex2f(cos(phi) * MINDIST, sin(phi) * MINDIST);
    phi := phi + 4*pi / 6
  until phi > 4 * pi + 0.1;
  glEnd;

  glEndList;
end;

procedure InitGL(Form: TForm);
begin
  Form.WindowState := wsMaximized;

  DC := GetDC(Form.Handle);
  SetDCPixelFormat(DC);
  hrc := wglCreateContext(DC);
  wglMakeCurrent(DC, hrc);
  glClearColor(0, 0, 0, 1);

  GLEnable(GL_BLEND);
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_POLYGON_SMOOTH);
  glEnable(GL_LINE_SMOOTH);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  ResizeGL(Form);

  CurGamma.A := 0;

  InitTextures;
  InitLists;
end;

procedure ResizeGL(Form: TForm);
begin
  glViewPort(0, 0, Form.ClientWidth, Form.ClientHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  glOrtho(0, MAXX * THEGAME.VIEWSCALE, MAXY * THEGAME.VIEWSCALE, 0, -1, 1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  //InvalidateRect(Form.Handle, nil, False);
end;

procedure FreeGL(Form: TForm);
begin
  glDeleteLists(1, 5);
  wglMakeCurrent(0, 0);
  wglDeleteContext(hrc);
  ReleaseDC(Form.Handle, DC);
  DeleteDC(DC);
end;

procedure UpdateFrame;
var
  I: integer;
  DY, R: GLfloat;
  E: TElement;
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  if Assigned(THEGAME.Player) then
  begin
    glPushMatrix;
    glTranslatef(
      -THEGAME.Player.X + MAXX * THEGAME.VIEWSCALE / 2,
      -THEGAME.Player.Y + MAXY * THEGAME.VIEWSCALE / 2,
      0);
    SetColor(CLR(FLOOR_COLOR, FLOOR_COLOR, FLOOR_COLOR, 1));
    TexturedRect(0, 0, MAXX, MAXY, TEX_GROUND, 0.2);
    THEGAME.Draw;
    glPopMatrix;
    DrawLine(0,
        10*THEGAME.VIEWSCALE,
        MAXX * THEGAME.VIEWSCALE,
        10*THEGAME.VIEWSCALE,
        BLACK, BLACK, WideActive);
    DrawLine(0,
        10*THEGAME.VIEWSCALE,
        (TARGETMONSTERS-THEGAME.MonstersRemains)/TARGETMONSTERS*MAXX * THEGAME.VIEWSCALE,
        10*THEGAME.VIEWSCALE,
        RED, GREEN, WideActive);
    DY := 0;
    R := 10*THEGAME.VIEWSCALE;
    for E := Low(TElement) to High(TElement) do
      for I := 1 to THEGAME.Player.Inventory[E] do
      begin
        glPushMatrix;
        DY := DY + R*2.1;
        glTranslatef(R*3, DY, 0);
        Circle(R, ElementCoreColor(E), ElementColor(E), CLR(0,0,0,-1));
        glPopMatrix;
      end;
  end;
  case THEGAME.Player.CurForm of
  pfWarrior: DrawRect(0,0,MAXX*THEGAME.VIEWSCALE, MAXY*THEGAME.VIEWSCALE, WARRIORSIGHT);
//  pfZombie: DrawRect(0,0,MAXX*THEGAME.VIEWSCALE, MAXY*THEGAME.VIEWSCALE, ZOMBIESIGHT);
  pfDemon: DrawRect(0,0,MAXX*THEGAME.VIEWSCALE, MAXY*THEGAME.VIEWSCALE, DEMONSIGHT);
  pfSpeed: DrawRect(0,0,MAXX*THEGAME.VIEWSCALE, MAXY*THEGAME.VIEWSCALE, HASTESIGHT);
  end;

  //Gamma shift:
  if CurGamma.A > 0 then
  begin
    DrawRect(0,0,MAXX*THEGAME.VIEWSCALE, MAXY*THEGAME.VIEWSCALE, CurGamma);
    CurGamma.A := CurGamma.A - GAMMA_DECR;
  end;
  SwapBuffers(DC);
end;

procedure DrawObject(L: integer; X, Y: GLfloat);
begin
  SetColor(WHITE);
  glPushMatrix;
  glTranslatef(X, Y, 0);
  glCallList(L);
  glPopMatrix;
end;

procedure DrawLine(X1, Y1, X2, Y2: GLfloat; Color1, Color2: TOpenGLColor;
  Effect: TLineEffect);
var
  I: integer;
  aR, aX, aY: GLfloat;
begin
  case Effect of
    Scratch:
    begin
      glLineWidth(2);
      Color1 := WHITE;
      Color2 := WHITE;
    end;
    Target:
    begin
      glLineWidth(1);
      Color1 := GREEN;
      Color2 := GREEN;
    end;
    Current:
    begin
      glLineWidth(2);
      Color2 := TRANSPARENT;
    end;
    Ready:
    begin
{      if (Color1 = Color2)and() then
      begin
        Color1 := MixColor(Color1, BLACK, 0.8);
        Color2 := MixColor(Color2, BLACK, 0.8);
      end;}
      glLineWidth(2);
    end;
    Active, WideActive:
    begin
      glLineWidth(1);
      glBegin(GL_POINTS);
             glBegin(GL_LINES);
      for I := 1 to 10 do
      begin
        aR := random;
        aX := X1 + aR * (X2 - X1);
        aY := Y1 + aR * (Y2 - Y1);
        SetColor(Color1);
        glVertex2d(aX + random * MINDIST - MINDIST / 2, aY + random * MINDIST - MINDIST / 2);
        SetColor(Color2);
        glVertex2d(aX + random * MINDIST - MINDIST / 2, aY + random * MINDIST - MINDIST / 2);
      end;
      glEnd;
{      glBegin(GL_LINES);
      aR := random;
      aX := X1 + aR * (X2 - X1);
      aY := Y1 + aR * (Y2 - Y1);
      SetColor(Color1);
      glVertex2d(aX + random * MINDIST - MINDIST / 2, aY + random * MINDIST - MINDIST / 2);
      SetColor(Color2);
      glVertex2d(aX + random * MINDIST - MINDIST / 2, aY + random * MINDIST - MINDIST / 2);
      glEnd;}
      if Effect = Active then
        glLineWidth(3)
      else if Effect = WideActive then
        glLineWidth(10)
    end;
  end;
  glBegin(GL_LINES);
  SetColor(Color1);
  glVertex2f(X1, Y1);
  SetColor(Color2);
  glVertex2f(X2, Y2);
  glEnd;
end;

procedure DrawRect(X1, Y1, X2, Y2: GLfloat; const Color: TOpenGLColor);
begin
  SetColor(Color);
  glBegin(GL_QUADS);
  glVertex2f(X2, Y2);
  glVertex2f(X2, Y1);
  glVertex2f(X1, Y1);
  glVertex2f(X1, Y2);
  glEnd;
end;

procedure TexturedRect(X1, Y1, X2, Y2: GLfloat; Tex: integer; TexScale: extended);
begin
  TexScale := TexScale / MINDIST;
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, Tex);
  glBegin(GL_QUADS);
  glTexCoord2f(0, 0);
  glVertex2f(X1, Y1);
  glTexCoord2f(0, (Y2 - Y1) * TexScale);
  glVertex2f(X1, Y2);
  glTexCoord2f((X2 - X1) * TexScale, (Y2 - Y1) * TexScale);
  glVertex2f(X2, Y2);
  glTexCoord2f((X2 - X1) * TexScale, 0);
  glVertex2f(X2, Y1);
  glEnd;
  glDisable(GL_TEXTURE_2D);
end;

procedure SetColor(const C: TOpenGLColor);
begin
  glColor4f(C.R, C.G, C.B, C.A);
end;

procedure SetRandomColor(const C, Delta: TOpenGLColor);
var
  E: GLfloat;
begin
  E := random;
  glColor4f(C.R + E * Delta.R, C.G + E * Delta.G, C.B + E * Delta.B, C.A + E * Delta.A);
end;

function MixColor(C1, C2: TOpenGLColor; Rate: GLfloat): TOpenGLColor;
begin
  Result.R := C1.R + Rate * (C2.R - C1.R);
  Result.G := C1.G + Rate * (C2.G - C1.G);
  Result.B := C1.B + Rate * (C2.B - C1.B);
  Result.A := C1.A + Rate * (C2.A - C1.A);
end;

procedure Circle(R: GLfloat; CenterColor, EdgeColor, VarColor: TOpenGLColor);
var
  phi: GLfloat;
begin
  SetColor(CenterColor);
  glbegin(GL_TRIANGLE_FAN);
  glVertex2f(0, 0);
  phi := 0;
  repeat
    SetRandomColor(EdgeColor, VarColor);
    glVertex2f(cos(phi) * R, sin(phi) * R);
    phi := phi + pi / 20
  until phi > 2 * pi + 0.1;
  glEnd;
end;

function CLR(aR, aG, aB, aA: GLfloat): TOpenGLColor;
begin
  Result.R := aR;
  Result.G := aG;
  Result.B := aB;
  Result.A := aA;
end;

function InvCLR(C: TOpenGLColor): TOpenGLColor;
begin
  Result.R := -C.R;
  Result.G := -C.R;
  Result.B := -C.B;
  Result.A := -C.A;
end;

operator = (C1, C2: TOpenGLColor): boolean;
begin
  Result := (C1.R = C2.R)and(C1.G = C2.G)and(C1.B = C2.B)and(C1.A = C2.A);
end;

procedure SetScreenGamma(C: TOpenGLColor);
begin
  CurGamma.R := C.R;
  CurGamma.G := C.G;
  CurGamma.B := C.B;
  CurGamma.A := C.A;
end;

end.

